// **************************************************************************
// * This file is part of the FreeFileSync project. It is distributed under *
// * GNU General Public License: http://www.gnu.org/licenses/gpl.html       *
// * Copyright (C) ZenJu (zhnmju123 AT gmx DOT de) - All Rights Reserved    *
// **************************************************************************

#include "unittest.h"
#include "../process_xml.h"
#include <wx/filefn.h>
#include "../ffs_paths.h"
#include "../parse_plural.h"

using namespace zen;


TEST(XmlInputOutput, GUI)
{
    //cleanup
    wxString testXml = wxT("test語.xml");
    if (wxFileExists(testXml))
        wxRemoveFile(testXml);

    ASSERT_FALSE(wxFileExists(testXml));

    //load structure with gui settings "guiCfg"
    xmlAccess::XmlGuiConfig guiCfg;
    guiCfg.mainCfg.cmpConfig.compareVar = CMP_BY_TIME_SIZE;
    guiCfg.mainCfg.syncCfg.directionCfg.var             = DirectionConfig::MIRROR;
    guiCfg.mainCfg.syncCfg.directionCfg.custom.exLeftSideOnly  = zen::SYNC_DIR_LEFT;
    guiCfg.mainCfg.syncCfg.directionCfg.custom.exRightSideOnly = zen::SYNC_DIR_NONE;
    guiCfg.mainCfg.syncCfg.directionCfg.custom.leftNewer       = zen::SYNC_DIR_LEFT;
    guiCfg.mainCfg.syncCfg.directionCfg.custom.rightNewer      = zen::SYNC_DIR_RIGHT;
    guiCfg.mainCfg.syncCfg.directionCfg.custom.different       = zen::SYNC_DIR_RIGHT;
    guiCfg.mainCfg.syncCfg.directionCfg.custom.conflict        = zen::SYNC_DIR_RIGHT;
    guiCfg.mainCfg.globalFilter.includeFilter = Zstr("include語\\");
    guiCfg.mainCfg.globalFilter.excludeFilter = Zstr("語exclude\\8");
    guiCfg.mainCfg.syncCfg.handleDeletion = zen::MOVE_TO_CUSTOM_DIRECTORY;
    guiCfg.handleError = xmlAccess::ON_GUIERROR_IGNORE;

    //attention: hidden settings currently are NOT persisted!
    //guiCfg.mainCfg.hidden

    guiCfg.mainCfg.firstPair =
        FolderPairEnh(Zstr("&left1語ojk"),
                      Zstr("&right1語iu"),
                      std::shared_ptr<CompConfig>(),
                      std::shared_ptr<SyncConfig>(),
                      FilterConfig());


    //add. pairs
    SyncConfig altCfg;
    altCfg.directionCfg.custom.exLeftSideOnly = SYNC_DIR_LEFT;
    altCfg.handleDeletion = DELETE_PERMANENTLY;
    altCfg.customDeletionDirectory = wxT("239d i90q dasf 9");

    guiCfg.mainCfg.additionalPairs.push_back(
        FolderPairEnh(Zstr("&left1語"),
                      Zstr("&right1語"),
                      std::shared_ptr<CompConfig>(),
                      std::shared_ptr<SyncConfig>(new SyncConfig(altCfg)),
                      FilterConfig()));

    FilterConfig altFilt;
    altFilt.includeFilter = Zstr("sdafi93423");
    altFilt.excludeFilter = Zstr("420opysdf");

    guiCfg.mainCfg.additionalPairs.push_back(
        FolderPairEnh(Zstr("国語1"),
                      Zstr("国語2"),
                      std::shared_ptr<CompConfig>(),
                      std::shared_ptr<SyncConfig>(),
                      FilterConfig(altFilt)));


    guiCfg.hideFilteredElements = false;
    guiCfg.handleError = xmlAccess::ON_GUIERROR_POPUP;
    guiCfg.syncPreviewEnabled   = false;

    //write XML
    ASSERT_NO_THROW(writeConfig(guiCfg, testXml));

    ASSERT_TRUE(xmlAccess::getXmlType(testXml) == xmlAccess::XML_TYPE_GUI);

    //read XML
    xmlAccess::XmlGuiConfig guiCfg2;
    ASSERT_NO_THROW(xmlAccess::readConfig(testXml, guiCfg2));

    //test if retrieved file is equal created file
    ASSERT_TRUE(guiCfg == guiCfg2);

    //clean up
    wxRemoveFile(testXml);
}


TEST(XmlInputOutput, Batch)
{
    //cleanup
    wxString testXml = wxT("test語.xml");
    if (wxFileExists(testXml))
        wxRemoveFile(testXml);

    ASSERT_FALSE(wxFileExists(testXml));

    //load structure with batch settings "batchCfg"
    xmlAccess::XmlBatchConfig batchCfg;
    batchCfg.mainCfg.cmpConfig.compareVar = CMP_BY_TIME_SIZE;
    batchCfg.mainCfg.syncCfg.directionCfg.var                    = DirectionConfig::UPDATE;
    batchCfg.mainCfg.syncCfg.directionCfg.custom.exLeftSideOnly  = zen::SYNC_DIR_LEFT;
    batchCfg.mainCfg.syncCfg.directionCfg.custom.exRightSideOnly = zen::SYNC_DIR_NONE;
    batchCfg.mainCfg.syncCfg.directionCfg.custom.leftNewer       = zen::SYNC_DIR_LEFT;
    batchCfg.mainCfg.syncCfg.directionCfg.custom.rightNewer      = zen::SYNC_DIR_RIGHT;
    batchCfg.mainCfg.syncCfg.directionCfg.custom.different       = zen::SYNC_DIR_RIGHT;
    batchCfg.mainCfg.globalFilter.includeFilter = Zstr("include語\\");
    batchCfg.mainCfg.globalFilter.excludeFilter = Zstr("語exclude\\8");
    batchCfg.mainCfg.syncCfg.handleDeletion = zen::MOVE_TO_CUSTOM_DIRECTORY;

    batchCfg.handleError = xmlAccess::ON_ERROR_IGNORE;
    batchCfg.mainCfg.firstPair =
        FolderPairEnh(Zstr("&left1語"),
                      Zstr("&right1語"),
                      std::shared_ptr<CompConfig>(),
                      std::shared_ptr<SyncConfig>(),
                      FilterConfig());

    batchCfg.showProgress = true;

    batchCfg.logFileDirectory = wxT("custom deletion dir");

    //write XML
    ASSERT_NO_THROW(writeConfig(batchCfg, testXml));

    ASSERT_TRUE(xmlAccess::getXmlType(testXml) == xmlAccess::XML_TYPE_BATCH);

    //read XML
    xmlAccess::XmlBatchConfig batchCfg2;
    ASSERT_NO_THROW(xmlAccess::readConfig(testXml, batchCfg2));

    //test if retrieved file is equal created file
    ASSERT_TRUE(batchCfg.mainCfg == batchCfg2.mainCfg);
    ASSERT_EQ(batchCfg.handleError, batchCfg2.handleError);
    ASSERT_EQ(batchCfg.showProgress, batchCfg2.showProgress);
    ASSERT_EQ(batchCfg.logFileDirectory, batchCfg2.logFileDirectory);

    //clean up
    wxRemoveFile(testXml);
}


inline
bool colAttrIsEqual(const ColumnAttributeRim& a, const ColumnAttributeRim& b)
{
    return (a.type_    == b.type_     &&
            a.visible_ == b.visible_  &&
            a.width_   == b.width_);
}


TEST(XmlInputOutput, GlobalSettings)
{
    //load structure with settings
    xmlAccess::XmlGlobalSettings globalCfg;

    globalCfg.programLanguage         = wxLANGUAGE_ICELANDIC;
    globalCfg.optDialogs.warningDependentFolders = false;

    globalCfg.gui.dlgSize.x            = 1111;
    globalCfg.gui.dlgSize.y            = 2222;
    globalCfg.gui.dlgPos.x             = 3333;
    globalCfg.gui.dlgPos.y             = 4444;
    globalCfg.gui.isMaximized          = true;

    auto& attrLeft = globalCfg.gui.columnAttribLeft;
    attrLeft.push_back(ColumnAttributeRim(COL_TYPE_DATE,      1132, false));
    attrLeft.push_back(ColumnAttributeRim(COL_TYPE_REL_PATH,  2040, true));
    attrLeft.push_back(ColumnAttributeRim(COL_TYPE_FULL_PATH, 2550, false));
    attrLeft.push_back(ColumnAttributeRim(COL_TYPE_DIRECTORY, 2600, true));
    attrLeft.push_back(ColumnAttributeRim(COL_TYPE_FILENAME,  1530, false));
    attrLeft.push_back(ColumnAttributeRim(COL_TYPE_SIZE,       8330, false));
    attrLeft.push_back(ColumnAttributeRim(COL_TYPE_EXTENSION,  620, true));

    auto& attrRight = globalCfg.gui.columnAttribRight;
    attrRight = attrLeft;
    attrRight.push_back(ColumnAttributeRim(COL_TYPE_DIRECTORY, 26040, false));
    attrRight.push_back(ColumnAttributeRim(COL_TYPE_FILENAME,  15430, true));
    attrRight.push_back(ColumnAttributeRim(COL_TYPE_SIZE,       48330, true));


    //write XML
    ASSERT_NO_THROW(writeConfig(globalCfg));

    ASSERT_TRUE(xmlAccess::getXmlType(xmlAccess::getGlobalConfigFile()) == xmlAccess::XML_TYPE_GLOBAL);

    //read XML
    xmlAccess::XmlGlobalSettings globalCfg2;
    ASSERT_NO_THROW(xmlAccess::readConfig(globalCfg2));

    //test if retrieved file is equal created file
    ASSERT_EQ(globalCfg.programLanguage, globalCfg2.programLanguage);
    ASSERT_EQ(globalCfg.optDialogs.warningDependentFolders, globalCfg2.optDialogs.warningDependentFolders);

    ASSERT_EQ(globalCfg.gui.dlgSize.x, globalCfg2.gui.dlgSize.x);
    ASSERT_EQ(globalCfg.gui.dlgSize.y, globalCfg2.gui.dlgSize.y);
    ASSERT_EQ(globalCfg.gui.dlgPos.x, globalCfg2.gui.dlgPos.x);
    ASSERT_EQ(globalCfg.gui.dlgPos.y, globalCfg2.gui.dlgPos.y);
    ASSERT_EQ(globalCfg.gui.isMaximized, globalCfg2.gui.isMaximized);
    ASSERT_EQ(globalCfg.gui.columnAttribLeft.size(), globalCfg2.gui.columnAttribLeft.size());
    ASSERT_EQ(globalCfg.gui.columnAttribRight.size(), globalCfg2.gui.columnAttribRight.size());
    ASSERT_TRUE(equal(globalCfg.gui.columnAttribLeft.begin(), globalCfg.gui.columnAttribLeft.end(), globalCfg2.gui.columnAttribLeft.begin(), colAttrIsEqual));
    ASSERT_TRUE(equal(globalCfg.gui.columnAttribRight.begin(), globalCfg.gui.columnAttribRight.end(), globalCfg2.gui.columnAttribRight.begin(), colAttrIsEqual));
}




const char formPol[] = " n == 1 ? 0 : n % 10 >= 2 && n % 10 <= 4 && (n % 100 < 10 || n % 100 > 20) ? 1 : 2"; //nplurals=3;
int tstPol(int n)
{
    return n == 1 ? 0 : n % 10 >= 2 && n % 10 <= 4 && (n % 100 < 10 || n % 100 > 20) ? 1 : 2;
}

const char formRu[] = "n % 10 == 1 && n % 100 != 11 ? 0 : n % 10 >= 2 && n % 10 <= 4 && (n % 100 < 10 || n % 100 > 20) ? 1 : 2"; //nplurals= 3
int tstRu(int n)
{
    return n % 10 == 1 && n % 100 != 11 ? 0 : n % 10 >= 2 && n % 10 <= 4 && (n % 100 < 10 || n % 100 > 20) ? 1 : 2 ;
}

const char formLit[] = " n% 10 == 1&& n % 100 != 11 ? 0 : n % 100 != 12&& n % 10 == 2 ? 1 : 0"; //nplurals =3;
int tstLit(int n)
{
    return n % 10 == 1 && n % 100 != 11 ? 0 : n % 100 != 12 && n % 10 == 2 ? 1 : 0;
}

const char formArab[] = "n == 0 ? 0 : n == 1 ? 1 : n == 2 ? 2 : (n % 100 >= 3 && n % 100 <= 10) ? 3 : (n % 100 >= 11 && n % 100 <= 99) || (n % 100 == 1) || (n % 100 ==2) ? 4 : 5"; //nplurals = 6
int tstArab(int n)
{
    return n == 0 ? 0 : n == 1 ? 1 : n == 2 ? 2 : (n % 100 >= 3 && n % 100 <= 10) ? 3 : (n % 100 >= 11 && n % 100 <= 99) || (n % 100 == 1) || (n % 100 == 2) ? 4 : 5;
}

const char formGerm[] = " n == 1 ? 0 : 1"; //nplurals=2
int tstGerm(int n)
{
    return n == 1 ? 0 : 1;
}

const char formFren[] = "n <= 1 ? 0 : 1"; //nplurals=2
int tstFren(int n)
{
    return n <= 1 ? 0 : 1;
}

const char formJap[] = " 0 "; //nplurals=1;
int tstJap(int n)
{
    return 0;
}

const char formRom[] = "n == 1 ? 0 : n == 0 || (n % 100 >= 1 && n % 100 <= 20) ? 1 : 2	"; //nplurals=3
int tstRom(int n)
{
    return n == 1 ? 0 : n == 0 || (n % 100 >= 1 && n % 100 <= 20) ? 1 : 2	;
}

const char formCze[] = "n % 100 == 1 ? 0 : n % 100 >= 2 && n % 100 <= 4 ? 1 : 2"; //nplurals=3;
int tstCze(int n)
{
    return n % 100 == 1 ? 0 : n % 100 >= 2 && n % 100 <= 4 ? 1 : 2;
}

const char formSlov[] = "   n%100==1 ? 0 : n%100==2 ? 1 : n%100==3 || n%100==4 ? 2 : 3 "; //nplurals=4
int tstSlov(int n)
{
    return n % 100 == 1 ? 0 : n % 100 == 2 ? 1 : n % 100 == 3 || n % 100 == 4 ? 2 : 3;
}

TEST(Misc, PluralFormsParser)
{
    typedef int (*TestFun)(int);
    typedef std::vector<std::pair<const char*, TestFun> > PhraseFunList;
    PhraseFunList phrases;
    phrases.push_back(std::make_pair(formPol,  &tstPol));
    phrases.push_back(std::make_pair(formRu,   &tstRu));
    phrases.push_back(std::make_pair(formLit,  &tstLit));
    phrases.push_back(std::make_pair(formArab, &tstArab));
    phrases.push_back(std::make_pair(formGerm, &tstGerm));
    phrases.push_back(std::make_pair(formFren, &tstFren));
    phrases.push_back(std::make_pair(formJap,  &tstJap));
    phrases.push_back(std::make_pair(formRom,  &tstRom));
    phrases.push_back(std::make_pair(formCze,  &tstCze));
    phrases.push_back(std::make_pair(formSlov, &tstSlov));

    for (PhraseFunList::const_iterator i = phrases.begin(); i != phrases.end(); ++i)
    {
        PluralForm pf(i->first);
        for (int j = 0; j < 1000000; ++j) //blunt, but relatively fast: ~2s
            ASSERT_EQ((i->second)(j), pf.getForm(j));
    }
}
