// **************************************************************************
// * This file is part of the FreeFileSync project. It is distributed under *
// * GNU General Public License: http://www.gnu.org/licenses/gpl.html       *
// * Copyright (C) ZenJu (zhnmju123 AT gmx DOT de) - All Rights Reserved    *
// **************************************************************************

#include <iostream>
#include <fstream>
#include <wx/wx.h>
#include <wx/dir.h>
#include <iterator>
#include <map>
#include <set>
#include <vector>
#include "../shared/parse_lng.h"

using namespace lngfile;

//reads all texts marked as _("dummy") into file %2 for translation (and preserve content including already translated lines)
//discards entries that are not imported newly via the files from directory %1



void normalize(std::string& sourceStr)
{
    std::string output;

    for (std::string::const_iterator iter = sourceStr.begin(); iter != sourceStr.end(); ++iter)
    {
        //read backslash
        if (*iter == '\\')
        {
            //read next character
            ++iter;
            if (iter == sourceStr.end())
            {
                output += '\\';
                break;
            }

            switch (*iter)
            {
                case '\\':
                    output += '\\';
                    break;
                case '\"':
                    output += '\"';
                    break;
                case 'n':
                    output += '\n';
                    break;
                case 't':
                    output += '\t';
                    break;
                default:
                {
                    output += '\\';
                    output += *iter;
                }
            }
        }
        else
            output += *iter;
    }

    sourceStr = output;
}


size_t findStrEnd(const std::string& line, size_t indexStart)
{
    size_t indexEnd = indexStart;
    for (;;)
    {
        indexEnd = line.find("\"", indexEnd);
        if (indexEnd == std::string::npos)
            return std::string::npos;

        if (indexEnd == 0)
            return 0;

        if (line[indexEnd - 1] != '\\')
            return indexEnd;

        ++indexEnd;
    }
}


void extractRegular(const std::string& stream, const std::set<std::string>& excludeList, const TranslationMap& transInput, TranslationList& transOutput)
{
    size_t textStart = 0;
    for (;;)
    {
        //------------------------------------------------------------
        textStart = stream.find("_(\"", textStart);
        if (textStart == std::string::npos)
            return;
        if (textStart != 0)
            if (isalpha(static_cast<unsigned char>(stream[textStart - 1]))) //exclude phrases like: version_("1.0")
            {
                textStart += 1;
                continue;
            }
        textStart += 3;

        size_t textEnd = findStrEnd(stream, textStart);
        if (textEnd == std::string::npos)
            return;

        std::string original = stream.substr(textStart, textEnd - textStart);
        textStart = textEnd + 1;
        //------------------------------------------------------------

        normalize(original); //resolve \" to "

        if (!original.empty()) //do not add empty rows
        {
            //add element only if NOT in excludeList
            if (excludeList.find(original) == excludeList.end())
            {
                std::string translation;

                //try to load entries from changing file (inOutFile)
                TranslationMap::const_iterator iter = transInput.find(original);
                if (iter != transInput.end())
                    translation = iter->second;

                transOutput.addItem(original, translation);
            }
        }
    }
}


void extractPlural(const std::string& stream, const TranslationPluralMap& transInputPlural, TranslationList& transOutput)
{
    size_t textStart = 0;
    for (;;)
    {
        //------------------------------------------------------------
        textStart = stream.find("_P(\"", textStart);
        if (textStart == std::string::npos)
            return;
        if (textStart != 0)
            if (isalpha(static_cast<unsigned char>(stream[textStart - 1]))) //exclude phrases like: version_("1.0")
            {
                textStart += 1;
                continue;
            }
        textStart += 4;

        size_t textEnd = findStrEnd(stream, textStart);
        if (textEnd == std::string::npos)
            return;

        std::string engSingular = stream.substr(textStart, textEnd - textStart);
        textStart = textEnd + 1;
        //------------------------------------------------------------

        textStart = stream.find("\"", textStart);
        if (textStart == std::string::npos)
            return;
        textStart += 1;

        textEnd = findStrEnd(stream, textStart);
        if (textEnd == std::string::npos)
            return;

        std::string engPlural = stream.substr(textStart, textEnd - textStart);
        textStart = textEnd + 1;
        //------------------------------------------------------------

        normalize(engSingular); //resolve \" to "
        normalize(engPlural);   //

        SingularPluralPair original(engSingular, engPlural);

        if (!original.first.empty() && !original.second.empty()) //do not add empty rows
        {
            PluralForms translation;

            //try to load entries from changing file (inOutFile)
            TranslationPluralMap::const_iterator iter = transInputPlural.find(original);
            if (iter != transInputPlural.end())
                translation = iter->second;

            transOutput.addPluralItem(original, translation);
        }
    }
}


int main(int argc, const char* argv[])
{
    if (argc != 3)
    {
        std::cout << "\n" << "Invalid input parameters: Specify input directory to search (%1) and changing language file (%2)" << "\n\n";
        system("pause");
        return -1;
    }

    const std::string inOutFile = argv[2];

    //read inOutFile into buffer, if existing
    TranslationMap       transInput;
    TranslationPluralMap transInputPlural;
    TransHeader          lngHeader;
    {
        std::ifstream input(inOutFile.c_str(), std::ios::binary);
        input >> std::noskipws;

        std::string inputStream((std::istream_iterator<char>(input)), std::istream_iterator<char>());

        try
        {
            parseLng(inputStream, lngHeader, transInput, transInputPlural); //throw ParsingError
        }
        catch (ParsingError& e)
        {
            std::cout << "Error parsing file " << inOutFile.c_str() << "!" << "\n";
            std::cout << "Row: "    << e.row << "\n";
            std::cout << "Column: " << e.col << "\n";
            system("pause");
            return -1;
        }
    }

    //read list of texts to exclude
    std::set<std::string> excludeList;
    std::ifstream exclude("exclude.txt");
    if (exclude)
    {
        std::vector<char> buffer(100000);
        while (exclude.getline(&buffer[0], buffer.size()))
            excludeList.insert(&buffer[0]);
    }

    //get input files (*.cpp and *.h files including sub-dirs of directory specified in arg[1]
    wxArrayString inputFiles;

    wxDir::GetAllFiles(wxString::FromUTF8(argv[1]), &inputFiles, wxT("*.h"));
    wxDir::GetAllFiles(wxString::FromUTF8(argv[1]), &inputFiles, wxT("*.cpp"));

    TranslationList transOutput; //keep translation in unsorted(!) list

    for (size_t i = 0; i < inputFiles.GetCount(); ++i)
    {
        std::ifstream input(inputFiles[i].ToUTF8(), std::ios::binary);
        if (input)
        {
            input >> std::noskipws;
            std::string inputStream((std::istream_iterator<char>(input)), std::istream_iterator<char>());

            extractRegular(inputStream, excludeList, transInput, transOutput);
            extractPlural (inputStream, transInputPlural, transOutput);
        }
        else
        {
            std::cout << "Unable to read " << inputFiles[i].ToUTF8() << "!" << "\n";
            system("pause");
            return -1;
        }
    }

    std::string outputStream;
    generateLng(transOutput, lngHeader, outputStream);

    std::ofstream output(inOutFile.c_str());
    output.write(outputStream.c_str(), outputStream.size());

    std::cout << "\nOperation successful! ";
    if (transOutput.untranslatedTextExists())
        std::cout << "\nTexts to be translated are existing!!!" << "\n\n";
    else
        std::cout << "All OK!" << "\n\n";

    return 0;
}
